import numpy as np
import readgadget
import redshift_space_library as RSL
import MAS_library as MASL
import units_library as UL
import Pk_library as PKL
import sys,os

########### routines ############
# Pk_comp(snapshot_fname,ptype,dims,do_RSD,axis,cpus)
# Pk_Gadget(snapshot_fname,dims,particle_type,do_RSD,axis,cpus)
#################################


U = UL.units();  rho_crit = U.rho_crit

# dictionary for files name
name_dict = {'0' :'GAS',  '01':'GCDM',  '02':'GNU',    '04':'Gstars',
             '1' :'CDM',                '12':'CDMNU',  '14':'CDMStars',
             '2' :'NU',                                '24':'NUStars',
             '4' :'Stars',
             '-1':'matter'}

###############################################################################
# This routine computes the power spectrum of either a single species or of all
# species from a Gadget routine
# snapshot_fname -----------> name of the Gadget snapshot
# ptype --------------------> scalar: 0-GAS, 1-CDM, 2-NU, 4-Stars, -1:ALL
# dims ---------------------> Total number of cells is dims^3 to compute Pk
# do_RSD -------------------> Pk in redshift-space (True) or real-space (False)
# axis ---------------------> axis along which move particles in redshift-space
# cpus ---------------------> Number of cpus to compute power spectra 
# folder_out ---------------> directory where to save the output
def Pk_comp(snapshot_fname,ptype,dims,do_RSD,axis,cpus,folder_out):

    # read relevant paramaters on the header
    print('Computing power spectrum...')
    head     = readgadget.header(snapshot_fname)
    BoxSize  = head.boxsize/1e3 #Mpc/h
    Masses   = head.massarr*1e10 #Msun/h
    Nall     = head.nall;  Ntotal = np.sum(Nall,dtype=np.int64)
    Omega_m  = head.omega_m
    Omega_l  = head.omega_l
    redshift = head.redshift
    Hubble   = 100.0*np.sqrt(Omega_m*(1.0+redshift)**3+Omega_l)  #km/s/(Mpc/h)
    z        = '%.3f'%redshift
        
    # find output file name
    fout = folder_out+'/Pk_' + name_dict[str(ptype)]
    if do_RSD:  fout += ('_RS_axis=' + str(axis) + '_z=' + z + '.dat')
    else:       fout +=                           ('_z=' + z + '.dat')

    # read the positions of the particles
    pos = readgadget.read_block(snapshot_fname,"POS ",[ptype])/1e3 #Mpc/h
    print('%.3f < X [Mpc/h] < %.3f'%(np.min(pos[:,0]),np.max(pos[:,0])))
    print('%.3f < Y [Mpc/h] < %.3f'%(np.min(pos[:,1]),np.max(pos[:,1])))
    print('%.3f < Z [Mpc/h] < %.3f\n'%(np.min(pos[:,2]),np.max(pos[:,2])))

    # read the velocities of the particles
    if do_RSD:
        print('moving particles to redshift-space...')
        vel = readgadget.read_block(snapshot_fname,"VEL ",[ptype]) #km/s
        RSL.pos_redshift_space(pos,vel,BoxSize,Hubble,redshift,axis)
        del vel;  print('done')

    # define delta array
    delta = np.zeros((dims,dims,dims),dtype=np.float32)

    # when dealing with all particles take into account their different masses
    if ptype==-1:
        if Nall[0]==0: #if not hydro
            M = np.zeros(Ntotal,dtype=np.float32) #define the mass array
            offset = 0
            for ptype in [0,1,2,3,4,5]:
                M[offset:offset+Nall[ptype]] = Masses[ptype]
                offset += Nall[ptype]
        else:
            M = readgadget.read_block(snapshot_fname,"MASS",ptype=[-1])*1e10
        
        mean = np.sum(M,dtype=np.float64)/dims**3
        MASL.MA(pos,delta,BoxSize,'CIC',M); del pos,M

    else:  
        mean = len(pos)*1.0/dims**3
        MASL.MA(pos,delta,BoxSize,'CIC'); del pos

    # compute the P(k) and save results to file
    delta /= mean;  delta -= 1.0
    Pk = PKL.Pk(delta,BoxSize,axis=axis,MAS='CIC',threads=cpus);  del delta
    np.savetxt(fout,np.transpose([Pk.k3D, Pk.Pk[:,0], Pk.Pk[:,1], Pk.Pk[:,2],
                                  Pk.Nmodes3D]))
###############################################################################

###############################################################################
# This routine computes the auto- and cross-power spectra of a Gadget snapshot
# in real or redshift-space. Can compute the total matter power spectrum or the
# auto- cross-power spectra of different particle types.
# If one only wants the total matter P(k), set particle_type=[-1]. If the P(k)
# of the different components is wanted set for instance particle_type=[0,1,4]
# snapshot_fname -----------> name of the Gadget snapshot
# dims ---------------------> Total number of cells is dims^3 to compute Pk
# particle_type ------------> compute Pk of those particles, e.g. [1,2]
# do_RSD -------------------> Pk in redshift-space (True) or real-space (False)
# axis ---------------------> axis along which move particles in redshift-space
# cpus ---------------------> Number of cpus to compute power spectra
# folder_out ---------------> folder where to put outputs
def Pk_Gadget(snapshot_fname,dims,particle_type,do_RSD,axis,cpus,
              folder_out=None):

    # find folder to place output files. Default is current directory
    if folder_out is None:  folder_out = os.getcwd()

    # for either one single species or all species use this routine
    if len(particle_type)==1:
        Pk_comp(snapshot_fname,particle_type[0],dims,do_RSD,
                axis,cpus,folder_out)
        return None

    # read snapshot head and obtain BoxSize, Omega_m and Omega_L
    print('\nREADING SNAPSHOTS PROPERTIES')
    head     = readgadget.header(snapshot_fname)
    BoxSize  = head.boxsize/1e3  #Mpc/h
    Nall     = head.nall
    Masses   = head.massarr*1e10 #Msun/h
    Omega_m  = head.omega_m
    Omega_l  = head.omega_l
    redshift = head.redshift
    Hubble   = 100.0*np.sqrt(Omega_m*(1.0+redshift)**3+Omega_l)  #km/s/(Mpc/h)
    h        = head.hubble
    z        = '%.3f'%redshift
    dims3    = dims**3

    # compute the values of Omega_cdm, Omega_nu, Omega_gas and Omega_s
    Omega_c = Masses[1]*Nall[1]/BoxSize**3/rho_crit
    Omega_n = Masses[2]*Nall[2]/BoxSize**3/rho_crit
    Omega_g, Omega_s = 0.0, 0.0
    if Nall[0]>0:
        if Masses[0]>0:  
            Omega_g = Masses[0]*Nall[0]/BoxSize**3/rho_crit
            Omega_s = Masses[4]*Nall[4]/BoxSize**3/rho_crit
        else:    
            # mass in Msun/h
            mass = readgadget.read_block(snapshot_fname,"MASS",ptype=[0])*1e10 
            Omega_g = np.sum(mass,dtype=np.float64)/BoxSize**3/rho_crit
            mass = readgadget.read_block(snapshot_fname,"MASS",ptype=[4])*1e10
            Omega_s = np.sum(mass,dtype=np.float64)/BoxSize**3/rho_crit
            del mass

    # some verbose
    print('Omega_gas    = ',Omega_g)
    print('Omega_cdm    = ',Omega_c)
    print('Omega_nu     = ',Omega_n)
    print('Omega_star   = ',Omega_s)
    print('Omega_m      = ',Omega_g + Omega_c + Omega_n + Omega_s)
    print('Omega_m snap = ',Omega_m)

    # dictionary giving the value of Omega for each component
    Omega_dict = {0:Omega_g, 1:Omega_c, 2:Omega_n, 4:Omega_s}
    #####################################################################

    # define the array containing the deltas
    delta = [[],[],[],[]]  #array containing the gas, CDM, NU and stars deltas

    # dictionary among particle type and the index in the delta and Pk arrays
    # delta of stars (ptype=4) is delta[3] not delta[4]
    index_dict = {0:0, 1:1, 2:2, 4:3} 

    # define suffix here
    if do_RSD:  suffix = '_RS_axis=' + str(axis) + '_z=' + z + '.dat'
    else:       suffix =                           '_z=' + z + '.dat'
    #####################################################################

    # do a loop over all particle types and compute the deltas
    for ptype in particle_type:
    
        # read particle positions in #Mpc/h
        pos = readgadget.read_block(snapshot_fname,"POS ",[ptype])/1e3 

        # move particle positions to redshift-space
        if do_RSD:
            vel = readgadget.read_block(snapshot_fname,"VEL ",[ptype])#km/s
            RSL.pos_redshift_space(pos,vel,BoxSize,Hubble,redshift,axis)
            del vel

        # find the index of the particle type in the delta array
        index = index_dict[ptype]

        # compute mean number of particles per grid cell
        mean_number = len(pos)*1.0/dims3

        # compute the deltas
        delta[index] = np.zeros((dims,dims,dims),dtype=np.float32)
        MASL.MA(pos,delta[index],BoxSize,'CIC');  del pos
        delta[index] /= mean_number;  delta[index] -= 1.0
    #####################################################################

    #####################################################################
    # if there are two or more particles compute auto- and cross-power spectra
    for i,ptype1 in enumerate(particle_type):
        for ptype2 in particle_type[i+1:]:

            # find the indexes of the particle types
            index1 = index_dict[ptype1];  index2 = index_dict[ptype2]

            # choose the name of the output files
            fout1  = '/Pk_' + name_dict[str(ptype1)]             + suffix
            fout2  = '/Pk_' + name_dict[str(ptype2)]             + suffix
            fout12 = '/Pk_' + name_dict[str(ptype1)+str(ptype2)] + suffix
            fout1  = folder_out + fout1
            fout2  = folder_out + fout2
            fout12 = folder_out + fout12

            # some verbose
            print('\nComputing the auto- and cross-power spectra of types: '\
                ,ptype1,'-',ptype2)
            print('saving results in:');  print (fout1,'\n',fout2,'\n',fout12)

            # This routine computes the auto- and cross-power spectra
            data = PKL.XPk([delta[index1],delta[index2]],BoxSize,axis=axis,
                           MAS=['CIC','CIC'],threads=cpus)
                                                        
            k = data.k3D;   Nmodes = data.Nmodes3D

            # save power spectra results in the output files
            np.savetxt(fout12,np.transpose([k,
                                            data.XPk[:,0,0],
                                            data.XPk[:,1,0],
                                            data.XPk[:,2,0],
                                            Nmodes]))
            np.savetxt(fout1, np.transpose([k,
                                            data.Pk[:,0,0],
                                            data.Pk[:,1,0],
                                            data.Pk[:,2,0],
                                            Nmodes]))
            np.savetxt(fout2, np.transpose([k,
                                            data.Pk[:,0,1],
                                            data.Pk[:,1,1],
                                            data.Pk[:,2,1],
                                            Nmodes]))
    #####################################################################

    #####################################################################
    # compute the power spectrum of the sum of all components
    print('\ncomputing P(k) of all components')

    # define delta of all components
    delta_tot = np.zeros((dims,dims,dims),dtype=np.float32)

    Omega_tot = 0.0;  fout = folder_out + '/Pk_'
    for ptype in particle_type:
        index = index_dict[ptype]
        delta_tot += (Omega_dict[ptype]*delta[index])
        Omega_tot += Omega_dict[ptype]
        fout += name_dict[str(ptype)] + '+'

    delta_tot /= Omega_tot;  del delta;  fout = fout[:-1] #avoid '+' in the end
    
    # compute power spectrum
    data = PKL.Pk(delta_tot,BoxSize,axis=axis,MAS='CIC',
                  threads=cpus);  del delta_tot

    # write P(k) to output file
    np.savetxt(fout+suffix, np.transpose([data.k3D,
                                          data.Pk[:,0],
                                          data.Pk[:,1],
                                          data.Pk[:,2],
                                          data.Nmodes3D]))
###############################################################################
###############################################################################



import numpy as np
import sys,os
import mass_function_library as MFL
import units_library as UL
import integration_library as IL

############################ CONSTANTS ############################
pi       = np.pi
rho_crit = (UL.units()).rho_crit #h^2 Msun/Mpc^3
deltac   = 1.686
###################################################################

##############################################################################
# This functions computes the halo bias, b(M), in a set of different halo masses
# k_in --------> input k
# Pk_in -------> input Pk
# OmegaM ------> value of Omega_m
# Masses ------> array with the halo masses where to evaluate the halo bias
# author ------> SMT01', 'Tinker'
def bias(k_in, Pk_in, OmegaM, Masses, author, bins=10000):

    rhoM = rho_crit*OmegaM

    # sort the input k-array
    indexes = np.argsort(k_in)
    k_in = k_in[indexes];  Pk_in = Pk_in[indexes]
    
    # interpolate the input k-array into bins points
    k  = np.logspace(np.log10(k_in[0]), np.log10(k_in[-1]), bins)
    Pk = np.interp(np.log(k), np.log(k_in), Pk_in)
    
    bias = np.zeros(Masses.shape[0], dtype=np.float64)

    if author=="SMT01":
        a = 0.707;  b = 0.5;  c = 0.6

        for i in range(Masses.shape[0]):
            R = (3.0*Masses[i]/(4.0*pi*rhoM))**(1.0/3.0)
            anu = a*(deltac/MFL.sigma(k,Pk,R))**2

            bias[i] = np.sqrt(a)*anu + np.sqrt(a)*b*anu**(1.0-c)
            bias[i] = bias[i] - anu**c/(anu**c + b*(1.0-c)*(1.0-0.5*c))
            bias[i] = 1.0+bias[i]/(np.sqrt(a)*deltac)        
    
    elif author=="Tinker":
        Delta=200.0;  y=np.log10(Delta)

        A=1.0+0.24*y*np.exp(-(4.0/y)**4);          a=0.44*y-0.88
        B=0.183;                                   b=1.5
        C=0.019+0.107*y+0.19*np.exp(-(4.0/y)**4);  c=2.4

        for i in range(Masses.shape[0]):
            R = (3.0*Masses[i]/(4.0*pi*rhoM))**(1.0/3.0)
            nu = deltac/MFL.sigma(k,Pk,R)
            bias[i] = 1.0-A*nu**a/(nu**a+deltac**a)+B*nu**b+C*nu**c

    return bias

##############################################################################
# This function returns the effective bias, see eq. 15 of Marulli 2011
def bias_eff(k, Pk, OmegaM, Masses, z, author):

    # compute halo mass function
    if author=='SMT01':
        dndM = MFL.MF_theory(k,Pk,OmegaM,Masses,'ST',bins=10000,z=z,delta=200.0)

    if author=='Tinker':
        dndM = MFL.MF_theory(k,Pk,OmegaM,Masses,'Tinker',bins=10000,z=z,delta=200.0)

    # compute halo bias
    b = bias(k, Pk, OmegaM, Masses, author, bins=10000)

    # integration parameters
    eps, h1, hmin = 1e-10, 1e0, 0.0

    yinit = np.array([0.0], dtype=np.float64)
    numerator   = IL.odeint_example2(yinit, np.min(Masses), np.max(Masses), eps,
                                     h1, hmin, Masses, dndM*b, verbose=False)[0]

    yinit = np.array([0.0], dtype=np.float64)
    denominator = IL.odeint_example2(yinit, np.min(Masses), np.max(Masses), eps,
                                     h1, hmin, Masses, dndM, verbose=False)[0]

    bias_eff = numerator/denominator

    return bias_eff








    






import numpy as np
import plotting_library as PL
import sys,os

from pylab import *
from matplotlib.ticker import ScalarFormatter
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1.inset_locator import mark_inset
from matplotlib.colors import LogNorm

################################# INPUT #######################################
# snapshot
snapshot = '/mnt/ceph/users/fvillaescusa/Quijote/Snapshots/latin_hypercube_HR/0/snapdir_004/snap_004'

# density field parameters
x_min, x_max = 0.0, 500.0
y_min, y_max = 0.0, 500.0
z_min, z_max = 0.0, 20.0
grid         = 1024
ptypes       = [1]   # 0-Gas, 1-CDM, 2-NU, 4-Stars; can deal with several species
plane        = 'XY'  #'XY','YZ' or 'XZ'
MAS          = 'PCS' #'NGP', 'CIC', 'TSC', 'PCS' 
save_df      = True  #whether save the density field into a file

# image parameters
fout            = 'Image.png'
min_overdensity = 0.5      #minimum overdensity to plot
max_overdensity = 50.0    #maximum overdensity to plot
scale           = 'log' #'linear' or 'log'
cmap            = 'hot'
###############################################################################

# compute 2D overdensity field
dx, x, dy, y, overdensity = PL.density_field_2D(snapshot, x_min, x_max, y_min, y_max,
                                             z_min, z_max, grid, ptypes, plane, MAS, save_df)


############### IMAGE ###############
print('\nCreating the figure...')
fig = figure()    #create the figure
ax1 = fig.add_subplot(111) 

ax1.set_xlim([x, x+dx])  #set the range for the x-axis
ax1.set_ylim([y, y+dy])  #set the range for the y-axis

ax1.set_xlabel(r'$h^{-1}{\rm Mpc}$',fontsize=18)  #x-axis label
ax1.set_ylabel(r'$h^{-1}{\rm Mpc}$',fontsize=18)  #y-axis label

if min_overdensity==None:  min_overdensity = np.min(overdensity)
if max_overdensity==None:  max_overdensity = np.max(overdensity)

overdensity[np.where(overdensity<min_overdensity)] = min_overdensity
overdensity[np.where(overdensity>max_overdensity)] = max_overdensity

if scale=='linear':
    cax = ax1.imshow(overdensity,cmap=get_cmap(cmap),origin='lower',
                     extent=[x, x+dx, y, y+dy], interpolation='bicubic',
                     vmin=min_overdensity,vmax=max_overdensity)
else:
    cax = ax1.imshow(overdensity,cmap=get_cmap(cmap),origin='lower',
                     extent=[x, x+dx, y, y+dy], interpolation='bicubic',
                     norm = LogNorm(vmin=min_overdensity,vmax=max_overdensity))

cbar = fig.colorbar(cax)
cbar.set_label(r"$\rho/\bar{\rho}$",fontsize=20)
savefig(fout, bbox_inches='tight')
close(fig)
#####################################




import numpy as np 
import time,sys,os
import pyfftw
import Pk_library as PKL
cimport numpy as np
cimport cython
#from cython.parallel import prange
from libc.math cimport sqrt,pow,sin,log10,abs
from libc.stdlib cimport malloc, free

 

# This function implement the MAS correction to modes amplitude
#@cython.cdivision(False)
#@cython.boundscheck(False)
cdef inline double MAS_correction(double x, int MAS_index):
    return (1.0 if (x==0.0) else pow(x/sin(x),MAS_index))

# This function computes the bispectrum of a triangle with k1, k2 and theta.
# theta can be an array and the output will be the bispectrum for the different
# values of theta
# delta --------> 3D density field
# BoxSize ------> size of the simulation box
# k1 -----------> wavenumber of k1
# k2 -----------> wavenumber of k2
# theta --------> array with the theta values. Set [theta0] for only one element
# MAS ----------> Mass assignment scheme used to compute density field
# threads ------> number of threads to compute the FFTs
@cython.boundscheck(False)
@cython.cdivision(False)
@cython.wraparound(False)
class Bk:
    def __init__(self, delta, BoxSize, k1, k2, theta, MAS='CIC', threads=1):
                 
        start = time.time()
        cdef int kxx, kyy, kzz, kx, ky, kz, MAS_index
        cdef int dims, middle, bins, i, j
        cdef unsigned long ID, dims2
        cdef list numbers
        cdef double k, prefact, pairs
        cdef double MAS_corr[3]
        cdef np.ndarray[np.float64_t, ndim=1] k_min, k_max
        cdef np.ndarray[np.float64_t, ndim=1] B, Q, triangles, kall, Pk
        ####### change this for double precision ######
        cdef float MAS_factor
        cdef np.complex64_t[:,:,::1] delta_k, delta1_k, delta2_k, delta3_k
        cdef np.complex64_t[:,:,::1] I1_k, I2_k, I3_k
        cdef np.float32_t[:,:,::1] delta1, delta2, delta3, I1, I2, I3
        ###############################################

        # find dimensions of delta: we assume is a (dims,dims,dims) array
        # determine the different frequencies and the MAS_index
        print('\nComputing bispectrum of the field...')
        dims = len(delta);  middle = dims//2;  dims2 = dims*dims
        kF,kN,kmax_par,kmax_per,kmax = PKL.frequencies(BoxSize,dims)
        MAS_index = PKL.MAS_function(MAS)

        # find the number of bins in theta. Define B, k_min, k_max arrays values of k3
        bins      = theta.shape[0]
        B         = np.zeros(bins,   dtype=np.float64)
        Q         = np.zeros(bins,   dtype=np.float64)
        triangles = np.zeros(bins,   dtype=np.float64)
        k_min     = np.zeros(bins+2, dtype=np.float64)
        k_max     = np.zeros(bins+2, dtype=np.float64)
        k_all     = np.zeros(bins+2, dtype=np.float64)
        Pk        = np.zeros(bins+2, dtype=np.float64)
        k3        = np.sqrt((k2*np.sin(theta))**2 + (k2*np.cos(theta)+k1)**2)
        k_all[0] = k1;  k_all[1] = k2;  k_all[2:] = k3

        # define the structure hosting the IDs of the cells within the spherical
        # shells. numbers[1] will be a list containing the IDs = dims^2*i+dims*j+k
        # of the cells that contribute to delta2_k. Find k_min and k_max for each k
        numbers = [];  numbers.append([]);  numbers.append([])
        k_min[0], k_max[0] = (k1-kF)/kF, (k1+kF)/kF
        k_min[1], k_max[1] = (k2-kF)/kF, (k2+kF)/kF
        for i in range(bins):  
            numbers.append([])
            k_min[i+2], k_max[i+2] = (k3[i]-kF)/kF, (k3[i]+kF)/kF
         
        ## compute FFT of the field (change this for double precision) ##
        delta_k = PKL.FFT3Dr_f(delta,threads)
        #################################

        # do a loop over the independent modes.
        # compute k,k_par,k_per, mu for each mode. k's are in kF units
        prefact = np.pi/dims
        for kxx in range(dims):
            kx = (kxx-dims if (kxx>middle) else kxx)
            MAS_corr[0] = MAS_correction(prefact*kx,MAS_index)
        
            for kyy in range(dims):
                ky = (kyy-dims if (kyy>middle) else kyy)
                MAS_corr[1] = MAS_correction(prefact*ky,MAS_index)

                for kzz in range(middle+1): #kzz=[0,1,..,middle] --> kz>0
                    kz = (kzz-dims if (kzz>middle) else kzz)
                    MAS_corr[2] = MAS_correction(prefact*kz,MAS_index)  

                    # kz=0 and kz=middle planes are special
                    """
                    if kz==0 or (kz==middle and dims%2==0):
                        if kx<0: continue
                        elif kx==0 or (kx==middle and dims%2==0):
                            if ky<0.0: continue
                    """

                    # correct modes amplitude for MAS
                    MAS_factor = MAS_corr[0]*MAS_corr[1]*MAS_corr[2]
                    delta_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]*MAS_factor

                    # find the ID of the cell
                    ID = dims2*kxx + dims*kyy + kzz
                    
                    # compute |k| of the mode and its integer part
                    k = sqrt(kx*kx + ky*ky + kz*kz)
                    for i in range(bins+2):
                        if k>=k_min[i] and k<k_max[i]:
                            numbers[i].append(ID)

        # fill the delta1_k array and compute delta1
        delta1_k = np.zeros((dims, dims, dims//2+1), dtype=np.complex64)
        I1_k     = np.zeros((dims, dims, dims//2+1), dtype=np.complex64)
        for i in range(len(numbers[0])):
            ID = numbers[0][i]
            kxx, kyy, kzz = ID//dims2, (ID%dims2)//dims, (ID%dims2)%dims
            delta1_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]
            I1_k[kxx,kyy,kzz]     = 1.0
        delta1 = PKL.IFFT3Dr_f(delta1_k, threads);  del delta1_k
        I1     = PKL.IFFT3Dr_f(I1_k,     threads);  del I1_k

        # compute Pk(k1)
        Pk[0],pairs = 0.0, 0.0
        for kxx in range(dims):        
            for kyy in range(dims):
                for kzz in range(dims):
                    Pk[0] += (delta1[kxx,kyy,kzz]*delta1[kxx,kyy,kzz])
                    pairs += (I1[kxx,kyy,kzz]*I1[kxx,kyy,kzz])
        Pk[0] = (Pk[0]/pairs)*(BoxSize/dims**2)**3

        # fill the delta2_k array and compute delta2
        delta2_k = np.zeros((dims, dims, dims//2+1), dtype=np.complex64)
        I2_k     = np.zeros((dims, dims, dims//2+1), dtype=np.complex64)
        for i in range(len(numbers[1])):
            ID = numbers[1][i]
            kxx, kyy, kzz = ID//dims2, (ID%dims2)//dims, (ID%dims2)%dims
            delta2_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]
            I2_k[kxx,kyy,kzz]     = 1.0
        delta2 = PKL.IFFT3Dr_f(delta2_k, threads);  del delta2_k
        I2     = PKL.IFFT3Dr_f(I2_k,     threads);  del I2_k
        
        # compute Pk(k2)
        Pk[1],pairs = 0.0, 0.0
        for kxx in range(dims):        
            for kyy in range(dims):
                for kzz in range(dims):
                    Pk[1] += (delta2[kxx,kyy,kzz]*delta2[kxx,kyy,kzz])
                    pairs += (I2[kxx,kyy,kzz]*I2[kxx,kyy,kzz])
        Pk[1] = (Pk[1]/pairs)*(BoxSize/dims**2)**3

        # fill the delta3_k array for the different theta bins
        for j in range(bins):
            delta3_k = np.zeros((dims, dims, dims//2+1), dtype=np.complex64)
            I3_k     = np.zeros((dims, dims, dims//2+1), dtype=np.complex64)
            for i in range(len(numbers[j+2])):
                ID = numbers[j+2][i]
                kxx, kyy, kzz = ID//dims2, (ID%dims2)//dims, (ID%dims2)%dims
                delta3_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]
                I3_k[kxx,kyy,kzz]     = 1.0
            delta3 = PKL.IFFT3Dr_f(delta3_k, threads);  del delta3_k
            I3     = PKL.IFFT3Dr_f(I3_k,     threads);  del I3_k

            # compute Pk(k3)
            Pk[j+2],pairs = 0.0, 0.0
            for kxx in range(dims):        
                for kyy in range(dims):
                    for kzz in range(dims):
                        Pk[j+2] += (delta3[kxx,kyy,kzz]*delta3[kxx,kyy,kzz])
                        pairs += (I3[kxx,kyy,kzz]*I3[kxx,kyy,kzz])
            Pk[j+2] = (Pk[j+2]/pairs)*(BoxSize/dims**2)**3

            # make the final sum and save bispectrum in B
            B[j] = 0.0
            for kxx in range(dims):        
                for kyy in range(dims):
                    for kzz in range(dims):
                        B[j] += (delta1[kxx,kyy,kzz]*delta2[kxx,kyy,kzz]*delta3[kxx,kyy,kzz])
                        triangles[j] += (I1[kxx,kyy,kzz]*I2[kxx,kyy,kzz]*I3[kxx,kyy,kzz])
            B[j] = (B[j]/triangles[j])*(BoxSize**2/dims**3)**3                
            Q[j] = B[j]/(Pk[0]*Pk[1]+Pk[0]*Pk[j+2]+Pk[1]*Pk[j+2])
    
        self.B         = B
        self.triangles = triangles
        self.Q         = Q
        self.k         = k_all
        self.Pk        = Pk
        print('Time to compute bispectrum = %.2f'%(time.time()-start))


# F2 kernel
def F2(k1_vec, k2_vec):
    k1_mod = np.sqrt(np.dot(k1_vec, k1_vec))
    k2_mod = np.sqrt(np.dot(k2_vec, k2_vec))
    ctheta = np.dot(k1_vec, k2_vec)/(k1_mod*k2_mod)
    return 5.0/7.0 + 1.0/2.0*ctheta*(k1_mod/k2_mod + k2_mod/k1_mod) + 2.0/7.0*ctheta**2
        

# This routine computes the tree-level bispectrum given k1,k2 and the linear Pk
def Bispectrum_theory(k,Pk,k1,k2):

    bins = 50
    B = np.zeros(bins, dtype=np.float64)

    thetas = np.linspace(0, np.pi, bins)
    k1_vec = np.array([0, 0, k1])

    Pk1 = np.interp(np.log(k1), np.log(k), Pk)
    Pk2 = np.interp(np.log(k2), np.log(k), Pk)

    for i,theta in enumerate(thetas):
        k2_vec = np.array([0,  k2*np.sin(theta), k2*np.cos(theta)])
        k3_vec = np.array([0, -k2*np.sin(theta),-k2*np.cos(theta)-k1])
        k3 = np.sqrt(np.dot(k3_vec, k3_vec))
        Pk3 = np.interp(np.log(k3), np.log(k), Pk)
        
        F2_12 = F2(k1_vec, k2_vec)
        F2_13 = F2(k1_vec, k3_vec)
        F2_23 = F2(k2_vec, k3_vec)

        B[i] = 2.0*Pk1*Pk2*F2_12 + 2.0*Pk1*Pk3*F2_13 + 2.0*Pk2*Pk3*F2_23
        print(F2_12, F2_13, F2_23)
        print(k1,k2,k3)
        print(Pk1,Pk2,Pk3)
        print(i,'\n')

    return thetas,B
        
        



cdef extern from "void_openmp_library.h":

      void mark_void_region(char *in_void, int Ncells, int dims, float R_grid2,
                            int i, int j, int k, int threads)

      int num_voids_around(long total_voids_found, int dims, 
                           float middle, int i, int j, int k, 
                           float *void_radius, int *void_pos, float R_grid, 
                           int threads)

      int num_voids_around2(int Ncells, int i, int j, int k, int dims, 
                            float R_grid2, char *in_void, int threads)



import numpy as np 
import time,sys,os,h5py,hdf5plugin
import readsnapHDF5 as rs
cimport numpy as np
cimport cython
from libc.math cimport sqrt,pow,sin,log10,abs,exp,log,rint
import readsnap#, groupcat
import units_library as UL
import MAS_library as MASL
import units_library as UL


################################# UNITS #####################################
cdef double rho_crit = (UL.units()).rho_crit 

cdef double yr    = 3.15576e7   #seconds
cdef double km    = 1e5         #cm
cdef double Mpc   = 3.0856e24   #cm
cdef double kpc   = 3.0856e21   #cm
cdef double Msun  = 1.989e33    #g
cdef double Ymass = 0.24        #helium mass fraction
cdef double mH    = 1.6726e-24  #proton mass in grams
cdef double gamma = 5.0/3.0     #ideal gas
cdef double kB    = 1.3806e-26  #gr (km/s)^2 K^{-1}
cdef double nu0   = 1420.0      #21-cm frequency in MHz
cdef double muH   = 2.3e-24     #gr
cdef double pi    = np.pi 
#############################################################################

# This routine computes the self-shielding parameters of Rahmati et al.
# and the amplitude of the UV background at a given redshift together with
def Rahmati_parameters(redshift, TREECOOL_file, Gamma_UVB=None, fac=1.0,
                       verbose=False):

    # Rahmati et. al. 2013 self-shielding parameters (table A1)
    z_t       = np.array([0.00, 1.00, 2.00, 3.00, 4.00, 5.00])
    n0_t      = np.array([-2.94,-2.29,-2.06,-2.13,-2.23,-2.35]); n0_t=10**n0_t
    alpha_1_t = np.array([-3.98,-2.94,-2.22,-1.99,-2.05,-2.63])
    alpha_2_t = np.array([-1.09,-0.90,-1.09,-0.88,-0.75,-0.57])
    beta_t    = np.array([1.29, 1.21, 1.75, 1.72, 1.93, 1.77])
    f_t       = np.array([0.01, 0.03, 0.03, 0.04, 0.02, 0.01])

    # compute the self-shielding parameters at the redshift of the N-body
    n0      = np.interp(redshift, z_t, n0_t)
    alpha_1 = np.interp(redshift, z_t, alpha_1_t)
    alpha_2 = np.interp(redshift, z_t, alpha_2_t)
    beta    = np.interp(redshift, z_t, beta_t)
    f       = np.interp(redshift, z_t, f_t)
    if verbose:
        print('n0 = %e\nalpha_1 = %2.3f\nalpha_2 = %2.3f\nbeta = %2.3f\n'\
            %(n0,alpha_1,alpha_2,beta) + 'f = %2.3f'%f)
        
    # find the value of the photoionization rate
    if Gamma_UVB is None:
        data = np.loadtxt(TREECOOL_file); logz=data[:,0]; Gamma_UVB=data[:,1]
        Gamma_UVB=np.interp(np.log10(1.0+redshift),logz,Gamma_UVB); del data
        if verbose:  print('Gamma_UVB(z=%2.2f) = %e s^{-1}'%(redshift,Gamma_UVB))
                     
    # Correct to reproduce the Lya forest mean flux
    Gamma_UVB /= fac

    return n0, alpha_1, alpha_2, beta, f, Gamma_UVB




# This routine computes the HI/H fraction of star-forming particles in the 
# Illustris-TNG simulation
# rho ---------------> star-forming part. densities in h^2 Msun/Mpc^3 (comoving)
# SPH ---------------> star-forming part radii in Mpc/h (comoving)
# metals ------------> star-forming part metallicities in solar units
# TREECOOL_file -----> file containing the UV strength
# Gamma -------------> value of the UVB if TREECOOL_file not provided
# fac ---------------> correction to the UVB to reproduce <F> of the Lya forest
# correct_H2 --------> whether to correct for H2
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def Rahmati_HI_Illustris(np.float32_t[:] rho, np.float32_t[:] SPH, 
                         np.float32_t[:] metals, redshift, h,
                         TREECOOL_file, Gamma=None, fac=1.0, 
                         correct_H2=False, verbose=False):
                         
    cdef long i, particles
    cdef float prefact, prefact2, nH, f_H2
    cdef double n0, alpha_1, alpha_2, beta, f, Gamma_UVB, chi, tau_c
    cdef double T, Lambda, alpha_A, Lambda_T, Gamma_phot, A, B, C, s
    cdef double sigma
    cdef np.float32_t[:] nH0

    # find the values of the self-shielding parameters and the UV background
    n0, alpha_1, alpha_2, beta, f, Gamma_UVB = \
        Rahmati_parameters(redshift, TREECOOL_file, Gamma, fac, verbose)
                      
    # find the number of star-forming particles and set their temperatures
    particles = rho.shape[0]
    T         = 1e4 #K

    # compute the case A recombination rate
    Lambda  = 315614.0/T
    alpha_A = 1.269e-13*pow(Lambda,1.503)\
        /pow(1.0 + pow((Lambda/0.522),0.47),1.923) #cm^3/s

    # Compute Lambda_T (eq. A6 of Rahmati et. al. 2013)
    Lambda_T = 1.17e-10*sqrt(T)*exp(-157809.0/T)/(1.0+sqrt(T/1e5)) #cm^3/s

    # prefactor to change densities from h^2 Msun/Mpc^3 to atoms/cm^{-3}
    prefact  = 0.76*h**2*Msun/Mpc**3/mH*(1.0+redshift)**3                   

    # prefactor to change surface densities from h Msun/Mpc^2 to g/cm^2
    prefact2 = (1.0+redshift)**2*h*Msun/Mpc**2 

    # define HI/H fraction
    nH0 = np.zeros(particles, dtype=np.float32)

    # do a loop over all particles
    for i in range(particles):

        # compute particle density in cm^{-3}. rho is in h^2 Msun/Mpc^3 units
        nH = rho[i]*prefact
        
        # compute the photoionization rate
        Gamma_phot = Gamma_UVB
        Gamma_phot *= ((1.0 - f)*pow(1.0 + pow(nH/n0,beta), alpha_1) +\
                           f*pow(1.0 + nH/n0, alpha_2))

        # compute the coeficients A,B and C to calculate the HI/H fraction
        A = alpha_A + Lambda_T
        B = 2.0*alpha_A + Gamma_phot/nH + Lambda_T
        C = alpha_A

        # compute the HI/H fraction (eq. A8 of Rahmati et. al. 2013)
        nH0[i] = (B-sqrt(B*B-4.0*A*C))/(2.0*A)

        ##### correct for H2 using KMT #####
        if correct_H2:
            chi   = 0.756*(1.0 + 3.1*metals[i]**0.365)           #dimensionless
            sigma = rho[i]*SPH[i]*prefact2                       #g/cm^2
            tau_c = sigma*(metals[i]*1e-21)/muH                  #dimensionless 
            s     = log(1.0 + 0.6*chi + 0.01*chi**2)/(0.6*tau_c) #dimensionless
            if s<2.0:  f_H2 = 1.0 - 0.75*s/(1.0 + 0.25*s)
            else:      f_H2 = 0.0
            if f_H2<0.0:  f_H2 = 0.0
            nH0[i] = nH0[i]*(1.0-f_H2)
            
    return np.asarray(nH0)

#########################################################################    
# This routine reads a single Illustris snapshot subfile and returns 
# particle positions and HI masses
cpdef HI_mass_from_Illustris_snap(snapshot, TREECOOL_file):

    f = h5py.File(snapshot, 'r')

    # read redshift and h
    redshift = f['Header'].attrs[u'Redshift']
    h        = f['Header'].attrs[u'HubbleParam']

    # read pos, radii, densities, HI/H and masses of gas particles 
    pos  = (f['PartType0/Coordinates'][:]/1e3).astype(np.float32)
    MHI  = f['PartType0/NeutralHydrogenAbundance'][:]
    mass = f['PartType0/Masses'][:]*1e10  #Msun/h
    rho  = f['PartType0/Density'][:]*1e19 #(Msun/h)/(Mpc/h)^3
    SFR  = f['PartType0/StarFormationRate'][:]
    indexes = np.where(SFR>0.0)[0];  del SFR
            
    # find the metallicity of star-forming particles
    metals = f['PartType0/GFM_Metallicity'][:]
    metals = metals[indexes]/0.0127

    # find densities of star-forming particles: units of h^2 Msun/Mpc^3
    Volume = mass/rho                            #(Mpc/h)^3
    radii  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 
    rho    = rho[indexes]                        #h^2 Msun/Mpc^3
    Volume = Volume[indexes]                     #(Mpc/h)^3

    # find volume and radius of star-forming particles
    radii_SFR  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 
        
    # find HI/H fraction for star-forming particles
    MHI[indexes] = Rahmati_HI_Illustris(rho, radii_SFR, metals, 
                                        redshift, h, TREECOOL_file, 
                                        Gamma=None, fac=1, 
                                        correct_H2=True) #HI/H
    MHI *= (0.76*mass)
    f.close()
    
    return pos, MHI

#########################################################################    
"""
# This routine identifies the offsets for halos and galaxies
# and perform several checks
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef test(f_offset, snapshot_root, num):

    cdef int i, j, Number
    cdef long offset, offset_halo, particles
    cdef np.int64_t[:] offset_galaxies, 
    cdef np.int64_t[:] end_halos, end_all_galaxies
    cdef np.int64_t[:] offset_halos, offset_subhalos
    cdef np.int32_t[:] lens_halos, subhalos_num, lens_subhalos

    # read halos and subhalos offset
    f = h5py.File(f_offset, "r")
    offset_halos    = f['Group/SnapByType'][:,0]
    offset_subhalos = f['Subhalo/SnapByType'][:,0]
    f.close()

    # read number of particles in halos and subhalos and number of subhalos
    halos = groupcat.loadHalos(snapshot_root, num, 
                               fields=['GroupLenType','GroupNsubs'])
    lens_halos   = halos['GroupLenType'][:,0]  
    subhalos_num = halos['GroupNsubs']
    subhalos = groupcat.loadSubhalos(snapshot_root, num, 
                                     fields=['SubhaloLenType','SubhaloMass'])
    lens_subhalos = subhalos['SubhaloLenType'][:,0]

    # define the array containing the beginning and ends of the halos
    end_halos = np.zeros(lens_halos.shape[0], dtype=np.int64)

    # find the offsets of the halos
    particles = 0
    for i in range(lens_halos.shape[0]):
        if offset_halos[i]!=particles:
            raise Exception('Offset are wrong!!')
        particles += lens_halos[i]
        end_halos[i] = particles
    del offset_halos


    # define the array hosting the offset of galaxies
    end_all_galaxies = np.zeros(lens_halos.shape[0], dtype=np.int64)
    offset_galaxies  = np.zeros(lens_subhalos.shape[0], dtype=np.int64)

    # do a loop over all halos
    Number, offset, offset_halo = 0, 0, 0
    for i in range(lens_halos.shape[0]):

        offset = offset_halo
        end_all_galaxies[i] = offset

        # do a loop over all galaxies belonging to the halo
        for j in range(subhalos_num[i]):
            offset_galaxies[Number] = offset
            end_all_galaxies[i] += lens_subhalos[Number]
            offset += lens_subhalos[Number];  Number += 1

        if (end_all_galaxies[i]-offset_halo)>lens_halos[i]:
            raise Exception('More particles in galaxies than in halo!!!')

        offset_halo += lens_halos[i]

    # check that galaxy offsets are the same
    for i in range(offset_galaxies.shape[0]):
        if offset_galaxies[i]!=offset_subhalos[i]:
            raise Exception('Offset of subhalos are different!!!')

    return np.asarray(end_halos),np.asarray(end_all_galaxies)
"""
#########################################################################    

# This routine computes the HI 
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef M_HI_counter(np.float32_t[:] NHI, np.float32_t[:] mass, 
                   np.float64_t[:] M_HI, np.float64_t[:] M_HI_gal, 
                   np.int64_t[:] end_halos, np.int64_t[:] end_all_galaxies, 
                   long Number, long start, long end, long end_gal,
                   long halo_num, done):

    cdef long j, particles, num_halos
    
    # find the number of particles to iterate over
    particles = NHI.shape[0]
    num_halos = M_HI.shape[0]

    # do a loop over all particles
    for j in range(particles):
        if Number>end:
            halo_num += 1
            if halo_num<num_halos:  
                start   = end
                end     = end_halos[halo_num]
                end_gal = end_all_galaxies[halo_num]
            else:
                done = True;  break

        # if particle is in galaxy add to M_HI_gal
        if Number>=start and Number<end_gal:
            M_HI_gal[halo_num] += 0.76*NHI[j]*mass[j]
                    
        M_HI[halo_num] += 0.76*NHI[j]*mass[j]
        Number += 1

    return Number, start, end, end_gal, halo_num, done
#######################################################################

#######################################################################
# This routine computes the HI mass within halos and the HI mass within
# halos that it is in galaxies
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef M_HI_halos_galaxies(np.int64_t[:] pars, done,
                          np.int32_t[:] halo_len, np.int32_t[:] gal_len,
                          np.int32_t[:] gal_in_halo,
                          np.float32_t[:] mass_ratio, np.float32_t[:] MHI,
                          np.float64_t[:] M_HI, np.float64_t[:] M_HI_gal):

    cdef long Number            = pars[0]
    cdef long start_h           = pars[1]
    cdef long end_h             = pars[2]
    cdef long start_g           = pars[3]
    cdef long end_g             = pars[4]
    cdef long halo_num          = pars[5]
    cdef long gal_num           = pars[6]
    cdef long gal_in_halo_local = pars[7]
    cdef long i, particles, num_halos, num_galaxies
    
    # find the number of particles to iterate over
    particles    = MHI.shape[0]
    num_halos    = M_HI.shape[0]
    num_galaxies = gal_len.shape[0]

    # do a loop over all particles
    for i in range(particles):

        # if particle belongs to new halo change variables
        if Number>=end_h:

            # accout for missing galaxies (should be galaxies with 0 particles)
            if Number>=end_g and gal_in_halo_local<gal_in_halo[halo_num]-1:
                gal_num += 1
                gal_in_halo_local += 1
                while gal_len[gal_num]==0 and \
                        gal_in_halo_local<gal_in_halo[halo_num]-1:
                    gal_num += 1
                    gal_in_halo_local += 1
            
            #if gal_num!=np.sum(gal_in_halo[:halo_num+1], dtype=np.int64)-1:
            #    print 'Numbers differ!!!!'
            #    print halo_num
            #    print gal_num
            #    print np.sum(gal_in_halo[:halo_num+1], dtype=np.int64)
                
            # update halo variables
            halo_num += 1

            if halo_num>=num_halos:
                # check that all galaxies have been counted
                if gal_num!=np.sum(gal_in_halo, dtype=np.int64)-1:
                    print('gal_num  = %ld'%gal_num)
                    print('galaxies = %ld'%(np.sum(gal_in_halo, dtype=np.int64)-1))
                    raise Exception("Finished without counting all galaxies")
                done = True;  break

            #if halo_num<num_halos:
                #start_h = end_h
                #end_h   = end_h + halo_len[halo_num]
            while halo_len[halo_num]==0 and halo_num<num_halos:
                gal_num  += gal_in_halo[halo_num]
                halo_num += 1
            start_h = end_h
            end_h   = end_h + halo_len[halo_num]

            # restart galaxy variables
            if gal_num<num_galaxies and gal_in_halo[halo_num]>0:
                gal_in_halo_local = 0
                gal_num += 1
                start_g = start_h
                end_g   = start_h + gal_len[gal_num]

        # if particle belongs to new galaxy change variables
        if Number>=end_g and gal_in_halo_local<gal_in_halo[halo_num]-1:
            gal_num += 1
            gal_in_halo_local += 1
            while gal_len[gal_num]==0 and \
                    gal_in_halo_local<gal_in_halo[halo_num]-1:
                gal_num += 1
                gal_in_halo_local += 1
            start_g = end_g
            end_g   = end_g + gal_len[gal_num]

        # if particle is inside galaxy add it to M_HI_gal
        if Number>=start_g and Number<end_g and mass_ratio[gal_num]>0.1:
            M_HI_gal[halo_num] += MHI[i]

        # for halos always add the HI mass
        M_HI[halo_num] += MHI[i]
        Number += 1

    pars[0], pars[1], pars[2]          = Number, start_h, end_h 
    pars[3], pars[4], pars[5], pars[6] = start_g, end_g, halo_num, gal_num
    pars[7]                            = gal_in_halo_local

    return done

#######################################################################
# This routine computes the HI mass within halos and the HI mass within
# halos that it is in galaxies
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef M_HI_halos_gal_cen_sat(np.int64_t[:] pars, done,
                             np.int32_t[:] halo_len, np.int32_t[:] gal_len,
                             np.int32_t[:] gal_in_halo,
                             np.float32_t[:] mass_ratio, np.float32_t[:] MHI,
                             np.float64_t[:] M_HI, 
                             np.float64_t[:] M_HI_gal,
                             np.float64_t[:] M_HI_cen, 
                             np.float64_t[:] M_HI_sat):

    cdef long Number            = pars[0]
    cdef long start_h           = pars[1]
    cdef long end_h             = pars[2]
    cdef long start_g           = pars[3]
    cdef long end_g             = pars[4]
    cdef long halo_num          = pars[5]
    cdef long gal_num           = pars[6]
    cdef long gal_in_halo_local = pars[7]
    cdef long i, particles, num_halos, num_galaxies
    
    # find the number of particles to iterate over
    particles    = MHI.shape[0]
    num_halos    = M_HI.shape[0]
    num_galaxies = gal_len.shape[0]

    # do a loop over all particles
    for i in range(particles):

        # if particle belongs to new halo change variables
        if Number>=end_h:

            # accout for missing galaxies (should be galaxies with 0 particles)
            if Number>=end_g and gal_in_halo_local<gal_in_halo[halo_num]-1:
                gal_num += 1
                gal_in_halo_local += 1
                while gal_len[gal_num]==0 and \
                        gal_in_halo_local<gal_in_halo[halo_num]-1:
                    gal_num += 1
                    gal_in_halo_local += 1
            
            #if gal_num!=np.sum(gal_in_halo[:halo_num+1], dtype=np.int64)-1:
            #    print 'Numbers differ!!!!'
            #    print halo_num
            #    print gal_num
            #    print np.sum(gal_in_halo[:halo_num+1], dtype=np.int64)
                
            # update halo variables
            halo_num += 1

            if halo_num>=num_halos:
                # check that all galaxies have been counted
                if gal_num!=np.sum(gal_in_halo, dtype=np.int64)-1:
                    print('gal_num  = %ld'%gal_num)
                    print('galaxies = %ld'%(np.sum(gal_in_halo, dtype=np.int64)-1))
                    raise Exception("Finished without counting all galaxies")
                done = True;  break

            #if halo_num<num_halos:
                #start_h = end_h
                #end_h   = end_h + halo_len[halo_num]
            while halo_len[halo_num]==0 and halo_num<num_halos:
                gal_num  += gal_in_halo[halo_num]
                halo_num += 1
            start_h = end_h
            end_h   = end_h + halo_len[halo_num]

            # restart galaxy variables
            if gal_num<num_galaxies and gal_in_halo[halo_num]>0:
                gal_in_halo_local = 0
                gal_num += 1
                start_g = start_h
                end_g   = start_h + gal_len[gal_num]

        # if particle belongs to new galaxy change variables
        if Number>=end_g and gal_in_halo_local<gal_in_halo[halo_num]-1:
            gal_num += 1
            gal_in_halo_local += 1
            while gal_len[gal_num]==0 and \
                    gal_in_halo_local<gal_in_halo[halo_num]-1:
                gal_num += 1
                gal_in_halo_local += 1
            start_g = end_g
            end_g   = end_g + gal_len[gal_num]

        # if particle is inside galaxy add it to M_HI_gal
        if Number>=start_g and Number<end_g and mass_ratio[gal_num]>0.1:
            M_HI_gal[halo_num] += MHI[i]
            if gal_in_halo_local==0:  M_HI_cen[halo_num] += MHI[i]
            else:                     M_HI_sat[halo_num] += MHI[i]

        # for halos always add the HI mass
        M_HI[halo_num] += MHI[i]
        Number += 1

    pars[0], pars[1], pars[2]          = Number, start_h, end_h 
    pars[3], pars[4], pars[5], pars[6] = start_g, end_g, halo_num, gal_num
    pars[7]                            = gal_in_halo_local

    return done

#######################################################################
# This routine computes the peculiar velocity of HI within halos
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef V_HI_halos(long[:] pars, done, int[:] halo_len, float[:,:] halo_vel,
                 float[:] MHI, float[:,:] vel,
                 double[:,::1] V_HI, double[:] M_HI):

    cdef long Number            = pars[0]
    cdef long start_h           = pars[1]
    cdef long end_h             = pars[2]
    cdef long halo_num          = pars[3]
    cdef long i, particles, num_halos
    cdef float dVx, dVy, dVz
    
    # find the number of particles to iterate over
    particles    = MHI.shape[0]
    num_halos    = M_HI.shape[0]

    # do a loop over all particles
    for i in range(particles):

        # if particle belongs to new halo change variables
        if Number>=end_h:
                
            # update halo variables
            halo_num += 1

            if halo_num>=num_halos:
                done = True;  break

            while halo_len[halo_num]==0 and halo_num<num_halos:
                halo_num += 1
            start_h = end_h
            end_h   = end_h + halo_len[halo_num]

        # for halos always add the HI mass
        M_HI[halo_num] += MHI[i]

        # compute (Vi-V_halo)^2*M_HIi
        V_HI[halo_num,0] += vel[i,0]*MHI[i]
        V_HI[halo_num,1] += vel[i,1]*MHI[i]
        V_HI[halo_num,2] += vel[i,2]*MHI[i]
        Number += 1

    pars[0], pars[1], pars[2], pars[3] = Number, start_h, end_h, halo_num
    return done
    
#######################################################################
# This routine computes the HI mass within halos and the HI velocity 
# dispersion within halos
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef sigma_HI_halos(long[:] pars, done,
                     int[:] halo_len, float[:,:] halo_vel,
                     float[:] MHI, float[:,:] vel,
                     double[:] sigma2_HI, double[:] M_HI):

    cdef long Number            = pars[0]
    cdef long start_h           = pars[1]
    cdef long end_h             = pars[2]
    cdef long halo_num          = pars[3]
    cdef long i, particles, num_halos
    cdef float dVx, dVy, dVz
    cdef double sigma2
    
    # find the number of particles to iterate over
    particles    = MHI.shape[0]
    num_halos    = M_HI.shape[0]

    # do a loop over all particles
    for i in range(particles):

        # if particle belongs to new halo change variables
        if Number>=end_h:
                
            # update halo variables
            halo_num += 1

            if halo_num>=num_halos:
                done = True;  break

            while halo_len[halo_num]==0 and halo_num<num_halos:
                halo_num += 1
            start_h = end_h
            end_h   = end_h + halo_len[halo_num]

        # for halos always add the HI mass
        M_HI[halo_num] += MHI[i]

        # compute (Vi-V_halo)^2*M_HIi
        dVx = vel[i,0]-halo_vel[halo_num,0]
        dVy = vel[i,1]-halo_vel[halo_num,1]
        dVz = vel[i,2]-halo_vel[halo_num,2]
        sigma2 = dVx*dVx + dVy*dVy + dVz*dVz
        sigma2_HI[halo_num] += (sigma2*MHI[i])

        Number += 1

    pars[0], pars[1], pars[2], pars[3] = Number, start_h, end_h, halo_num
    return done

###############################################################
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef HI_image(np.float32_t[:,:] pos, float x_min, float x_max, 
               float y_min, float y_max, float z_min, float z_max):

    start = time.time()
    cdef long i, particles
    cdef list indexes_list

    indexes_list = []
    particles = pos.shape[0]
    for i in range(particles):

        if pos[i,0]>=x_min and pos[i,0]<=x_max:
            if pos[i,1]>=y_min and pos[i,1]<=y_max:
                if pos[i,2]>=z_min and pos[i,2]<=z_max:
                    indexes_list.append(i)
                    
    print('Time taken = %.2f seconds'%(time.time()-start))
    return indexes_list


# This routine implements reads Gadget gas output and correct HI/H fractions
# to account for self-shielding and H2
# snapshot_fname -------> name of the N-body snapshot
# fac ------------------> factor to reproduce the mean Lya flux
# TREECOOL_file --------> TREECOOL file used in the N-body
# Gamma_UVB ------------> value of the UVB photoionization rate
# correct_H2 -----------> correct the HI/H fraction to account for H2
# if Gamma_UVB is set to None the value of the photoionization rate will be read
# from the TREECOOL file, otherwise it is used the Gamma_UVB value
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def Rahmati_HI_assignment(snapshot_fname, fac, TREECOOL_file, Gamma_UVB=None,
                          correct_H2=False, IDs=None, verbose=False):

    # read snapshot head and obtain BoxSize, Omega_m and Omega_L
    print('\nREADING SNAPSHOT PROPERTIES')
    head     = readsnap.snapshot_header(snapshot_fname)
    BoxSize  = head.boxsize/1e3 #Mpc/h
    Nall     = head.nall
    redshift = head.redshift

    # find the values of the self-shielding parameters and the UV background
    n0, alpha_1, alpha_2, beta, f, Gamma_UVB = \
        Rahmati_parameters(redshift, TREECOOL_file, Gamma_UVB, fac, verbose)

    #compute the HI/H fraction 
    nH0 = HI_from_UVB(snapshot_fname, Gamma_UVB, True, correct_H2,
                      f, n0, alpha_1, alpha_2, beta, SF_temperature=1e4)
    
    #read the gas masses
    mass = readsnap.read_block(snapshot_fname,"MASS",parttype=0)*1e10 #Msun/h

    #create the array M_HI and fill it
    M_HI = 0.76*mass*nH0;  del nH0,mass
    print('Omega_HI = %e'%(np.sum(M_HI,dtype=np.float64)/BoxSize**3/rho_crit))

    return M_HI



#This function compute the HI/H fraction given the photoionization rate and 
#the density of the gas particles
#snapshot_fname ------------> name of the N-body snapshot
#Gamma_UVB -----------------> value of the UVB photoionization rate
#self_shielding_correction -> apply (True) or not (False) the self-shielding
#correct_H2 ----------------> correct the HI masses to account for H2
#f -------------------------> parameter of the Rahmati et al model (see eq. A1)
#n0 ------------------------> parameter of the Rahmati et al model (see eq. A1)
#alpha_1 -------------------> parameter of the Rahmati et al model (see eq. A1)
#alpha_2 -------------------> parameter of the Rahmati et al model (see eq. A1)
#beta ----------------------> parameter of the Rahmati et al model (see eq. A1)
#SF_temperature ------------> associate temperature of star forming particles
#If SF_temperature is set to None it will compute the temperature of the SF
#particles from their density and internal energy
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def HI_from_UVB(snapshot_fname, double Gamma_UVB,
                self_shielding_correction=False,correct_H2=False,
                double f=0.0, double n0=0.0, double alpha_1=0.0,
                double alpha_2=0.0, double beta=0.0,
                double SF_temperature=0.0):

    cdef long gas_part,i 
    cdef double mean_mol_weight, yhelium, A, B, C, R_surf, P, prefact2
    cdef double nH, Lambda, alpha_A, Lambda_T, Gamma_phot, prefact, T
    cdef np.float32_t[:] rho,U,ne,SFR,nH0

    # read snapshot head and obtain BoxSize, Omega_m and Omega_L
    head     = readsnap.snapshot_header(snapshot_fname)
    BoxSize  = head.boxsize/1e3 #Mpc/h
    Nall     = head.nall
    Masses   = head.massarr*1e10 #Msun/h
    Omega_m  = head.omega_m
    Omega_l  = head.omega_l
    redshift = head.redshift
    Hubble   = 100.0*np.sqrt(Omega_m*(1.0+redshift)**3+Omega_l)  #h*km/s/Mpc
    h        = head.hubble
    gas_part = Nall[0]

    # read density, internal energy, electron fraction and star formation rate
    # density units: h^2 Msun / Mpc^3
    yhelium = (1.0-0.76)/(4.0*0.76) 
    U   = readsnap.read_block(snapshot_fname,"U   ",parttype=0) #(km/s)^2
    ne  = readsnap.read_block(snapshot_fname,"NE  ",parttype=0) #electron frac
    rho = readsnap.read_block(snapshot_fname,"RHO ",parttype=0)*1e10/1e-9 
    SFR = readsnap.read_block(snapshot_fname,"SFR ",parttype=0) #SFR

    # define HI/H fraction
    nH0 = np.zeros(gas_part, dtype=np.float32)

    #### Now compute the HI/H fraction following Rahmati et. al. 2013 ####
    prefact  = 0.76*h**2*Msun/Mpc**3/mH*(1.0+redshift)**3 
    prefact2 = (gamma-1.0)*(1.0+redshift)**3*h**2*Msun/kpc**3/kB 

    print('doing loop...');  start = time.time()    
    for i in range(gas_part):

        # compute particle density in cm^{-3}. rho is in h^2 Msun/Mpc^3 units
        nH = rho[i]*prefact

        # compute particle temperature
        if SFR[i]>0.0:
            T = SF_temperature
            P = prefact2*U[i]*rho[i]*1e-9  #K/cm^3
        else:
            mean_mol_weight = (1.0+4.0*yhelium)/(1.0+yhelium+ne[i])
            T = U[i]*(gamma-1.0)*mH*mean_mol_weight/kB
        
        #compute the case A recombination rate
        Lambda  = 315614.0/T
        alpha_A = 1.269e-13*pow(Lambda,1.503)\
                  /pow(1.0 + pow((Lambda/0.522),0.47),1.923) #cm^3/s

        #Compute Lambda_T (eq. A6 of Rahmati et. al. 2013)
        Lambda_T = 1.17e-10*sqrt(T)*\
                   exp(-157809.0/T)/(1.0+sqrt(T/1e5)) #cm^3/s

        #compute the photoionization rate
        Gamma_phot = Gamma_UVB
        if self_shielding_correction:
            Gamma_phot *= ((1.0 - f)*pow(1.0 + pow(nH/n0,beta), alpha_1) +\
                           f*pow(1.0 + nH/n0, alpha_2))

        #compute the coeficients A,B and C to calculate the HI/H fraction
        A = alpha_A + Lambda_T
        B = 2.0*alpha_A + Gamma_phot/nH + Lambda_T
        C = alpha_A

        #compute the HI/H fraction (eq. A8 of Rahmati et. al. 2013)
        nH0[i] = (B-sqrt(B*B-4.0*A*C))/(2.0*A)

        # correct for the presence of H2
        if correct_H2 and SFR[i]>0.0:
            R_surf = (P/1.7e4)**0.8
            nH0[i] = nH0[i]/(1.0 + R_surf)

    print('Time taken = %.3f s'%(time.time()-start))
    return np.asarray(nH0)
############################################################################


############################################################################
# This routine looks for halos around the selected halos
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef locate_nearby_halos(np.float32_t[:,:] halos_pos, 
        np.float32_t[:,:] selected_halos_pos, float BoxSize):

    print('Starting routine...');  start = time.time()
    cdef np.int64_t[:] index, index_sort, indexes, cube_indexes
    cdef int ii, jj, kk, dims2, iii, jjj, kkk
    cdef long all_halos, selected_halos, i, j, dims3
    cdef int dims, number
    cdef np.float32_t[:,:] halos_pos_new
    cdef float dist, x, y, z
    cdef np.int64_t[:] offset

    # find the total number of halos in the catalogue
    all_halos = halos_pos.shape[0]
    selected_halos = selected_halos_pos.shape[0]
    dims  = <int>(BoxSize/2.0) # we consider the max radius of a halo to be 2 Mpc/h
    dims2 = dims*dims
    dims3 = dims*dims*dims

    # for each halo compute its index
    index      = np.zeros(all_halos,   dtype=np.int64)-1 #initialize to -1
    index_sort = np.zeros(all_halos,   dtype=np.int64)
    cube_indexes = np.zeros(27, dtype=np.int64)

    # for each halo find its index
    for i in range(all_halos):
        ii = <int>(halos_pos[i,0]*dims/BoxSize)
        if ii==dims:  ii = 0
        jj = <int>(halos_pos[i,1]*dims/BoxSize)
        if jj==dims:  jj = 0
        kk = <int>(halos_pos[i,2]*dims/BoxSize)
        if kk==dims:  kk = 0

        index[i] = ii*dims2 + jj*dims + kk


    indexes = np.argsort(index)
    halos_pos_new = np.copy(halos_pos)
    print(index)

    for i in range(indexes.shape[0]):
        halos_pos_new[i,0] = halos_pos[indexes[i],0]
        halos_pos_new[i,1] = halos_pos[indexes[i],1]
        halos_pos_new[i,2] = halos_pos[indexes[i],2]
        index_sort[i]      = index[indexes[i]]


    # current_number = index_sort[0]
    # offset[index_sort[0]] = 0
    # for i in range(all_halos):
    #     if index_sort[i]!=current_number:
    #         offset[index_sort[i]] = i
    #         current_number = index_sort[i]
    # offset[all_halos] = dims3

    # current_number = dims3
    # for i in range(all_halos,-1, -1):
    #     if offset[i]==-1:  offset[i] = current_number
    #     else:              current_number = offset[i]


    # print index_sort

    for i in range(all_halos):
        if i%1000==0:
            print(i)
        x = halos_pos[i,0]
        y = halos_pos[i,1]
        z = halos_pos[i,2]
        for j in range(selected_halos):
            dist = sqrt((selected_halos_pos[j,0]-x)**2 + \
                        (selected_halos_pos[j,1]-y)**2 + \
                        (selected_halos_pos[j,2]-z)**2)

    # # do a loop over the selected halos
    # for i in range(selected_halos_pos.shape[0]):
    #     ii = <int>(selected_halos_pos[i,0]*dims/BoxSize)
    #     if ii==dims:  ii = 0
    #     jj = <int>(selected_halos_pos[i,1]*dims/BoxSize)
    #     if jj==dims:  jj = 0
    #     kk = <int>(selected_halos_pos[i,2]*dims/BoxSize)
    #     if kk==dims:  kk = 0
        
    #     number = 0
    #     for iii in range(ii-1, ii+2):
    #         for jjj in range(jj-1, jj+2):
    #             for kkk in range(kk-1, kk+2):
    #                 cube_indexes[number] = 

    #     cube_indexes

    #     for j in range(all_halos):

    #         if index

    print('Time taken = %.3f seconds'%(time.time()-start))


############################################################################
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef HI_profile(np.float32_t[:,:] halo_pos, np.float32_t[:] halo_R,
    np.int64_t[:] halo_id, np.float32_t[:,:] pos, np.float32_t[:] MHI, 
    np.int64_t[:] offset, np.float64_t[:,:] HI_mass_shell, 
    np.int64_t[:] part_in_halo, float BoxSize, float R1):
    
    start = time.time()
    cdef long i, j, particles, halos, index, index_box, begin, end
    cdef float x_h, y_h, z_h, dx, dy, dz, factor
    cdef float middle, R2, radius, radius2, R12
    cdef int dims, dims2, bins, bin, index_x, index_y, index_z, 
    cdef int ii, jj, kk, iii, jjj, kkk, index_x_min, index_x_max
    cdef int index_y_min, index_y_max, index_z_min, index_z_max

    particles = pos.shape[0]
    halos     = halo_pos.shape[0]
    bins      = HI_mass_shell.shape[1]
    dims      = <int>rint((offset.shape[0]-1)**(1.0/3.0))
    dims2     = dims*dims
    R12       = R1*R1
    middle    = BoxSize/2.0
    factor    = dims*1.0/BoxSize

    # do a loop over all halos
    for i in range(halos):

        x_h     = halo_pos[i,0]
        y_h     = halo_pos[i,1]
        z_h     = halo_pos[i,2]
        radius  = halo_R[i]
        radius2 = radius*radius
        #index   = halo_id[i]

        # in C, <int>(3.2)=3, but <int>(-0.3)=0, so add dims to avoid this
        index_x_min = <int>((x_h-radius)*factor+dims)
        index_x_max = <int>((x_h+radius)*factor+dims)
        index_y_min = <int>((y_h-radius)*factor+dims)
        index_y_max = <int>((y_h+radius)*factor+dims)
        index_z_min = <int>((z_h-radius)*factor+dims)
        index_z_max = <int>((z_h+radius)*factor+dims)

        #index_x = index/dims2
        #index_y = (index%dims2)/dims
        #index_z = (index%dims2)%dims

        #for ii in range(-1,2):
            #iii = (index_x + ii + dims)%dims

            #for jj in range(-1,2):
            #    jjj = (index_y + jj + dims)%dims

                 #for kk in range(-1,2):
                 #   kkk = (index_z + kk + dims)%dims
        for ii in range(index_x_min, index_x_max+1):
            iii = ii%dims

            for jj in range(index_y_min, index_y_max+1):
                jjj = jj%dims

                for kk in range(index_z_min, index_z_max+1):
                    kkk = kk%dims

                    index_box = dims2*iii + dims*jjj + kkk

                    begin = offset[index_box]
                    end   = offset[index_box+1]

                    for j in range(begin,end):

                        #x        = pos[j,0]
                        #y        = pos[j,1]
                        #z        = pos[j,2]
                        #MHI_part = MHI[j]

                        dx = abs(pos[j,0] - x_h)
                        if dx>middle:  dx = BoxSize - dx
                        if dx>radius:  continue

                        dy = abs(pos[j,1] - y_h)
                        if dy>middle:  dy = BoxSize - dy
                        if dy>radius:  continue

                        dz = abs(pos[j,2] - z_h)
                        if dz>middle:  dz = BoxSize - dz
                        if dz>radius:  continue

                        R2 = dx*dx + dy*dy + dz*dz
                        if R2<radius2:

                            if R2<=R12:  bin = 0
                            else:     
                                bin = <int>(log10(R2/R12)/log10(radius2/R12)*(bins-1))+1
                            HI_mass_shell[i,bin]  += MHI[j]
                            part_in_halo[i] += 1


    print('Time taken = %.2f seconds'%(time.time()-start))
######################################################################################

# This routine computes the HI profile of one signale halo using brute force
# This is used mainly for validation purposes
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef HI_profile_1halo_brute_force(np.float32_t[:] halo_pos, float radius,
    np.float32_t[:,:] pos, np.float32_t[:] MHI, 
    np.float64_t[:] HI_mass_shell, float BoxSize, float R1, np.int64_t[:] part_in_halo):
    
    
    start = time.time()
    cdef long j, particles
    cdef float x, y, z, x_h, y_h, z_h, dx, dy, dz
    cdef float MHI_part, middle, R2, radius2, R12
    cdef int bin, bins

    particles = pos.shape[0]
    bins      = HI_mass_shell.shape[0]
    R12       = R1*R1
    middle    = BoxSize/2.0

    x_h     = halo_pos[0]
    y_h     = halo_pos[1]
    z_h     = halo_pos[2]
    radius2 = radius*radius

    # do a loop over all particles
    for j in range(particles):

        x        = pos[j,0]
        y        = pos[j,1]
        z        = pos[j,2]
        MHI_part = MHI[j]

        dx = abs(x - x_h)
        if dx>middle:  dx = BoxSize - dx
        if dx>radius:  continue

        dy = abs(y - y_h)
        if dy>middle:  dy = BoxSize - dy
        if dy>radius:  continue

        dz = abs(z - z_h)
        if dz>middle:  dz = BoxSize - dz
        if dz>radius:  continue

        R2 = dx*dx + dy*dy + dz*dz
        if R2<radius2:

            if R2<=R12:  bin = 0
            else:     
                bin = <int>(log10(R2/R12)/log10(radius2/R12)*(bins-1))+1
            HI_mass_shell[bin]  += MHI_part
            part_in_halo[0] += 1


    print('Time taken = %.2f seconds'%(time.time()-start))



cpdef overlap_particles_halos(np.float32_t[:,:] halo_pos, float Rvir, float BoxSize,
        float x_min, float x_max, float y_min, float y_max, 
        float z_min, float z_max):

    cdef long halos, i
    cdef float x, y, z, x1, x2, y1, y2, z1, z2

    halos = halo_pos.shape[0]

    for i in range(halos):
        x = halo_pos[i,0]
        y = halo_pos[i,1]
        z = halo_pos[i,2]

        x1 = (x-Rvir)%BoxSize
        x2 = (x+Rvir)%BoxSize
        if x1<x_min or x1>x_max or x2<x_min or x2>x_max:
            return True

        y1 = (y-Rvir)%BoxSize
        y2 = (y+Rvir)%BoxSize
        if y1<y_min or y1>y_max or y2<y_min or y2>y_max:
            return True

        z1 = (z-Rvir)%BoxSize
        z2 = (z+Rvir)%BoxSize
        if z1<z_min or z1>z_max or z2<z_min or z2>z_max:
            return True

    return False

############################################################################
# This routine is used to compute the total HI mass within the radius of 
# dark matter halos.
# halo_pos ----------> sorted positions of the halos
# halo_R ------------> corresponding radii of the halos
# pos ---------------> sorted positions of the particles
# M_HI --------------> corresponding HI masses of the particles
# offset ------------> offset array with the begin,end information of each cell
# M_HI_halo ---------> array containing the total HI mass inside each halo
# BoxSize -----------> simulation box size. Needed for boundary conditions
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef void HI_mass_SO(np.float32_t[:,::1] halo_pos, np.float32_t[::1] halo_R,
    np.float32_t[:,::1] pos, np.float32_t[::1] MHI, 
    np.int64_t[::1] offset, np.float64_t[::1] M_HI_halo, float BoxSize):
    
    start = time.time()
    cdef long i, j, particles, halos, index, index_box, begin, end
    cdef float x_h, y_h, z_h, dx, dy, dz, factor
    cdef float middle, R2, radius, radius2
    cdef int dims, dims2, index_x_min, index_x_max
    cdef int index_y_min, index_y_max, index_z_min, index_z_max
    cdef int ii, jj, kk, iii, jjj, kkk

    particles = pos.shape[0]
    halos     = halo_pos.shape[0]
    dims      = <int>rint((offset.shape[0]-1)**(1.0/3.0))
    dims2     = dims*dims
    middle    = BoxSize/2.0
    factor    = dims*1.0/BoxSize

    # do a loop over all halos
    for i in range(halos):

        x_h     = halo_pos[i,0]
        y_h     = halo_pos[i,1]
        z_h     = halo_pos[i,2]
        radius  = halo_R[i]
        radius2 = radius*radius

        # in C, <int>(3.2)=3, but <int>(-0.3)=0, so add dims to avoid this
        index_x_min = <int>((x_h-radius)*factor+dims)
        index_x_max = <int>((x_h+radius)*factor+dims)
        index_y_min = <int>((y_h-radius)*factor+dims)
        index_y_max = <int>((y_h+radius)*factor+dims)
        index_z_min = <int>((z_h-radius)*factor+dims)
        index_z_max = <int>((z_h+radius)*factor+dims)

        # do a loop over the cell hosting the halo and its neighburgs
        for ii in range(index_x_min, index_x_max+1):
            iii = ii%dims

            for jj in range(index_y_min, index_y_max+1):
                jjj = jj%dims

                for kk in range(index_z_min, index_z_max+1):
                    kkk = kk%dims

                    index_box = dims2*iii + dims*jjj + kkk

                    # find the location of the particles belonging to the cell
                    # in the sorted array
                    begin = offset[index_box]
                    end   = offset[index_box+1]

                    # do a loop over all particles in the cell
                    for j in range(begin,end):

                        dx = abs(pos[j,0] - x_h)
                        if dx>middle:  dx = BoxSize - dx
                        if dx>radius:  continue

                        dy = abs(pos[j,1] - y_h)
                        if dy>middle:  dy = BoxSize - dy
                        if dy>radius:  continue

                        dz = abs(pos[j,2] - z_h)
                        if dz>middle:  dz = BoxSize - dz
                        if dz>radius:  continue

                        R2 = dx*dx + dy*dy + dz*dz
                        if R2<radius2:
                            M_HI_halo[i] += MHI[j]


    print('Time taken = %.2f seconds'%(time.time()-start))

# # This routine computes the cross-section of the DLAs
# @cython.boundscheck(False)
# @cython.cdivision(True)
# @cython.wraparound(False)
# def DLAs_cross_section_original(snapshot_root, snapnum, TREECOOL_file, resolution):

#     cdef int subfile_num
#     cdef long i, num_halos, begin, end, offset, start_h, end_h
#     # cdef np.float32_t[:,:] pos_h
#     # cdef np.float32_t[:] radii_h, MHI_h
#     cdef np.float64_t[:] M_HI_tot, cross_section
#     cdef np.int32_t[:] halo_len
#     # cdef np.float32_t[:] mass,SFR,metals,rho,Volume,

#     # find snapshot name and read header
#     snapshot = snapshot_root + 'snapdir_%03d/snap_%03d'%(snapnum, snapnum)
#     header   = rs.snapshot_header(snapshot)
#     redshift = header.redshift
#     BoxSize  = header.boxsize/1e3 #Mpc/h
#     filenum  = header.filenum
#     Omega_m  = header.omega0
#     Omega_L  = header.omegaL
#     h        = header.hubble

#     print '\nBoxSize         = %.1f Mpc/h'%BoxSize
#     print 'Number of files = %d'%filenum
#     print 'Omega_m         = %.3f'%Omega_m
#     print 'Omega_l         = %.3f'%Omega_L
#     print 'redshift        = %.3f'%redshift

#     # read number of particles in halos, their positions and masses
#     halos = groupcat.loadHalos(snapshot_root, snapnum, 
#             fields=['GroupLenType','GroupPos','GroupMass','Group_R_TopHat200'])
#     halo_len    = halos['GroupLenType'][:,0]  
#     halo_pos    = halos['GroupPos']/1e3
#     halo_mass   = halos['GroupMass']*1e10
#     halo_radius = halos['Group_R_TopHat200']/1e3
#     del halos

#     # factor to convert (Mpc/h)^{-2} to cm^{-2}
#     factor = h*(1.0+redshift)**2*\
#         (UL.units().Msun_g)/(UL.units().mH_g)/(UL.units().Mpc_cm)**2

#     # find the total number of halos
#     num_halos = halo_pos.shape[0]
#     print 'Found %d halos'%num_halos
#     M_HI_tot      = np.zeros(num_halos, dtype=np.float64)
#     cross_section = np.zeros(num_halos, dtype=np.float64)

#     # read the first subfile of the snapshot
#     subfile_num = 0
#     offset = 0
#     snapshot = snapshot_root + 'snapdir_%03d/snap_%03d.%d'\
#         %(snapnum, snapnum, subfile_num)
#     header = rs.snapshot_header(snapshot)
#     npart  = header.npart[0]

#     # find positions, HI masses and radii of the gas particles
#     pos  = rs.read_block(snapshot, 'POS ', parttype=0, verbose=False)/1e3
#     pos  = pos.astype(np.float32)
#     MHI  = rs.read_block(snapshot, 'NH  ', parttype=0, verbose=False)#HI/H
#     mass = rs.read_block(snapshot, 'MASS', parttype=0, verbose=False)*1e10
#     SFR  = rs.read_block(snapshot, 'SFR ', parttype=0, verbose=False)
#     indexes = np.where(SFR>0.0)[0];  del SFR

#     # find the metallicity of star-forming particles
#     metals = rs.read_block(snapshot, 'GZ  ', parttype=0, verbose=False)
#     metals = metals[indexes]/0.0127

#     # find densities of star-forming particles: units of h^2 Msun/Mpc^3
#     rho = rs.read_block(snapshot, 'RHO ', parttype=0, verbose=False)*1e19
#     Volume = mass/rho                            #(Mpc/h)^3
#     radii  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 

#     # find density and radius of star-forming particles
#     radii_SFR = radii[indexes]    
#     rho       = rho[indexes]

#     # find HI/H fraction for star-forming particles
#     MHI[indexes] = Rahmati_HI_Illustris(rho, radii_SFR, metals, redshift, 
#                                             h, TREECOOL_file, Gamma=None,
#                                             fac=1, correct_H2=True) #HI/H
#     MHI *= (0.76*mass)

#     # do a loop over all halos
#     begin_abs = 0
#     for i in range(num_halos):

#         # print i,subfile_num

#         # find where the halo starts and ends in the file
#         if i>0:  begin_abs += halo_len[i-1]
#         end     = begin_abs + halo_len[i]
#         start_h, end_h = 0, end-begin_abs
#         begin   = begin_abs

#         # if i==294689:
#         #     print begin_abs,end

#         # define arrays hosting positions, masses and radii of particles in halo
#         pos_h   = np.zeros((end-begin,3), dtype=np.float32)
#         radii_h = np.zeros(end-begin,     dtype=np.float32)
#         MHI_h   = np.zeros(end-begin,     dtype=np.float32)

#         # if halo is in current subfile fill it 
#         if end<=(offset+npart):
#             pos_h   = pos[begin-offset:end-offset]
#             radii_h = radii[begin-offset:end-offset]
#             MHI_h   = MHI[begin-offset:end-offset]
#             M_HI_tot[i] = np.sum(MHI_h, dtype=np.float64)

#             if len(pos_h)==0:  continue

#             # check if all particles can produce a single DLA
#             rho_h   = 3.0*MHI_h/(4.0*np.pi*radii_h**3)
#             NHI_max = factor*2.0*np.sum(rho_h*radii_h, dtype=np.float64)
#             if NHI_max<10**(20.3):  continue

#             if halo_radius[i]>0.0:
#                 dims = int(halo_radius[i]*1.0/resolution)
#             else:
#                 dims = 50

#             area = DLAs_cross_section_1halo(pos_h, radii_h, MHI_h, dims, 
#                 BoxSize, h, redshift)
#             cross_section[i] = area
#             if area>0.1:
#                 print i,area
#                 np.savetxt('borrar.dat',np.transpose([pos_h[:,0], pos_h[:,1], pos_h[:,2], radii_h, MHI_h]))  
#                 sys.exit()
#             continue
#             # f=open('cross_section.dat','a')
#             # f.write(str(halo_mass[i])+' '+str(area)+'\n')
#             # f.close()

#         else:
#             pos_h[0:offset+npart-begin]   = pos[begin-offset:]
#             radii_h[0:offset+npart-begin] = radii[begin-offset:]
#             MHI_h[0:offset+npart-begin]   = MHI[begin-offset:]
#             start_h = offset+npart-begin
#             begin = offset+npart

#         # do a loop over subfiles until all particles in the halo has been read
#         done = False
#         while not(done):
#             offset += npart;  subfile_num += 1;  print subfile_num,i
#             snapshot = snapshot_root + 'snapdir_%03d/snap_%03d.%d'\
#                 %(snapnum, snapnum, subfile_num)
#             header = rs.snapshot_header(snapshot)
#             npart  = header.npart[0]

#             # find positions, HI masses and radii of the gas particles
#             pos  = rs.read_block(snapshot, 'POS ', parttype=0, verbose=False)/1e3
#             pos  = pos.astype(np.float32)
#             MHI  = rs.read_block(snapshot, 'NH  ', parttype=0, verbose=False)#HI/H
#             mass = rs.read_block(snapshot, 'MASS', parttype=0, verbose=False)*1e10
#             SFR  = rs.read_block(snapshot, 'SFR ', parttype=0, verbose=False)
#             indexes = np.where(SFR>0.0)[0];  del SFR

#             # find the metallicity of star-forming particles
#             metals = rs.read_block(snapshot, 'GZ  ', parttype=0, verbose=False)
#             metals = metals[indexes]/0.0127

#             # find densities of star-forming particles: units of h^2 Msun/Mpc^3
#             rho = rs.read_block(snapshot, 'RHO ', parttype=0, verbose=False)*1e19
#             Volume = mass/rho                            #(Mpc/h)^3
#             radii  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 

#             # find density and radius of star-forming particles
#             radii_SFR = radii[indexes]    
#             rho       = rho[indexes]

#             # find HI/H fraction for star-forming particles
#             MHI[indexes] = Rahmati_HI_Illustris(rho, radii_SFR, metals, redshift, 
#                                                     h, TREECOOL_file, Gamma=None,
#                                                     fac=1, correct_H2=True) #HI/H
#             MHI *= (0.76*mass)

#             if end<=(offset+npart):
#                 pos_h[start_h:]   = pos[begin-offset:end-offset]
#                 radii_h[start_h:] = radii[begin-offset:end-offset]
#                 MHI_h[start_h:]   = MHI[begin-offset:end-offset]
#                 done = True
#                 M_HI_tot[i] = np.sum(MHI_h, dtype=np.float64)

#             else:
#                 pos_h[start_h:start_h+offset+npart-begin]   = pos[begin-offset:]
#                 radii_h[start_h:start_h+offset+npart-begin] = radii[begin-offset:]
#                 MHI_h[start_h:start_h+offset+npart-begin]   = MHI[begin-offset:]
#                 start_h = start_h+offset+npart-begin
#                 begin = offset+npart

#     return M_HI_tot, cross_section

"""
# This routine computes the cross-section of the DLAs
# resolution -----> spatial resolution in Mpc/h for the grid
# NHI_values -----> computes the cross-section for different NHI values
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def DLAs_cross_section(snapshot_root, snapnum, TREECOOL_file, resolution,
    NHI_values):

    cdef int subfile_num
    cdef long i, k, num_halos, begin, end, offset, start_h, end_h
    # cdef np.float32_t[:,:] pos_h
    # cdef np.float32_t[:] radii_h, MHI_h
    cdef np.float64_t[:] M_HI_tot, area 
    cdef np.int32_t[:] halo_len
    # cdef np.float64_t[:,:] cross_section
    # cdef np.float32_t[:] mass,SFR,metals,rho,Volume,

    # find snapshot name and read header
    snapshot = snapshot_root + 'snapdir_%03d/snap_%03d'%(snapnum, snapnum)
    header   = rs.snapshot_header(snapshot)
    redshift = header.redshift
    BoxSize  = header.boxsize/1e3 #Mpc/h
    filenum  = header.filenum
    Omega_m  = header.omega0
    Omega_L  = header.omegaL
    h        = header.hubble

    print('\nBoxSize         = %.1f Mpc/h'%BoxSize)
    print('Number of files = %d'%filenum)
    print('Omega_m         = %.3f'%Omega_m)
    print('Omega_l         = %.3f'%Omega_L)
    print('redshift        = %.3f'%redshift)

    # read number of particles in halos, their positions and masses
    halos = groupcat.loadHalos(snapshot_root, snapnum, 
            fields=['GroupLenType','GroupPos','GroupMass','Group_R_TopHat200'])
    halo_len    = halos['GroupLenType'][:,0]  
    halo_pos    = halos['GroupPos']/1e3
    halo_mass   = halos['GroupMass']*1e10
    halo_radius = halos['Group_R_TopHat200']/1e3
    del halos

    # factor to convert (Mpc/h)^{-2} to cm^{-2}
    factor = h*(1.0+redshift)**2*\
        (UL.units().Msun_g)/(UL.units().mH_g)/(UL.units().Mpc_cm)**2

    # find the total number of halos and define arrays contining MHI and sigma
    num_halos = halo_pos.shape[0]
    NHI_bins  = NHI_values.shape[0]
    print('Found %d halos'%num_halos)
    M_HI_tot      = np.zeros(num_halos,             dtype=np.float64)
    cross_section = np.zeros((num_halos, NHI_bins), dtype=np.float64)

    # do a loop over all halos
    subfile_num, offset, npart, begin_abs = -1, 0, 0, 0
    pos   = np.empty((0,3), dtype=np.float32)
    radii = np.empty(0,     dtype=np.float32)
    MHI   = np.empty(0,     dtype=np.float32)
    for i in range(num_halos):

        # find where the halo starts and ends in the file
        if i>0:  begin_abs += halo_len[i-1]
        end     = begin_abs + halo_len[i]
        start_h, end_h = 0, end-begin_abs
        begin   = begin_abs

        # define arrays hosting positions, masses and radii of particles in halo
        pos_h   = np.zeros((end-begin,3), dtype=np.float32)
        radii_h = np.zeros(end-begin,     dtype=np.float32)
        MHI_h   = np.zeros(end-begin,     dtype=np.float32)

        # if halo is in current subfile fill it 
        if end<=(offset+npart):
            pos_h   = pos[begin-offset:end-offset]
            radii_h = radii[begin-offset:end-offset]
            MHI_h   = MHI[begin-offset:end-offset]
            M_HI_tot[i] = np.sum(MHI_h, dtype=np.float64)

            if len(pos_h)==0:  continue

            # check if all particles can produce a single DLA
            rho_h   = 3.0*MHI_h/(4.0*np.pi*radii_h**3)
            NHI_max = factor*2.0*np.sum(rho_h*radii_h, dtype=np.float64)
            if NHI_max<10**(20.3):  continue

            if halo_radius[i]>0.0:  dims = int(halo_radius[i]*1.0/resolution)
            else:                   dims = 50

            cross_section[i] = DLAs_cross_section_1halo(pos_h, radii_h, MHI_h, dims, 
                BoxSize, h, redshift, NHI_values, NHI_bins)
            continue

        else:
            pos_h[0:offset+npart-begin]   = pos[begin-offset:]
            radii_h[0:offset+npart-begin] = radii[begin-offset:]
            MHI_h[0:offset+npart-begin]   = MHI[begin-offset:]
            start_h = offset+npart-begin
            begin = offset+npart

        # do a loop over subfiles until all particles in the halo has been read
        done = False
        while not(done):
            offset += npart;  subfile_num += 1;  print(subfile_num,i)
            snapshot = snapshot_root + 'snapdir_%03d/snap_%03d.%d'\
                %(snapnum, snapnum, subfile_num)
            header = rs.snapshot_header(snapshot)
            npart  = header.npart[0]

            # find positions, HI masses and radii of the gas particles
            pos  = rs.read_block(snapshot, 'POS ', parttype=0, verbose=False)/1e3
            pos  = pos.astype(np.float32)
            MHI  = rs.read_block(snapshot, 'NH  ', parttype=0, verbose=False)#HI/H
            mass = rs.read_block(snapshot, 'MASS', parttype=0, verbose=False)*1e10
            SFR  = rs.read_block(snapshot, 'SFR ', parttype=0, verbose=False)
            indexes = np.where(SFR>0.0)[0];  del SFR

            # find the metallicity of star-forming particles
            metals = rs.read_block(snapshot, 'GZ  ', parttype=0, verbose=False)
            metals = metals[indexes]/0.0127

            # find densities of star-forming particles: units of h^2 Msun/Mpc^3
            rho = rs.read_block(snapshot, 'RHO ', parttype=0, verbose=False)*1e19
            Volume = mass/rho                            #(Mpc/h)^3
            radii  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 

            # find density and radius of star-forming particles
            radii_SFR = radii[indexes]    
            rho       = rho[indexes]

            # find HI/H fraction for star-forming particles
            MHI[indexes] = Rahmati_HI_Illustris(rho, radii_SFR, metals, redshift, 
                                                    h, TREECOOL_file, Gamma=None,
                                                    fac=1, correct_H2=True) #HI/H
            MHI *= (0.76*mass)

            if end<=(offset+npart):
                pos_h[start_h:]   = pos[begin-offset:end-offset]
                radii_h[start_h:] = radii[begin-offset:end-offset]
                MHI_h[start_h:]   = MHI[begin-offset:end-offset]
                done = True
                M_HI_tot[i] = np.sum(MHI_h, dtype=np.float64)

                if len(pos_h)==0:  continue

                # check if all particles can produce a single DLA
                rho_h   = 3.0*MHI_h/(4.0*np.pi*radii_h**3)
                NHI_max = factor*2.0*np.sum(rho_h*radii_h, dtype=np.float64)
                if NHI_max<10**(20.3):  continue

                if halo_radius[i]>0.0:  dims = int(halo_radius[i]*1.0/resolution)
                else:                   dims = 50

                cross_section[i] = DLAs_cross_section_1halo(pos_h, radii_h, MHI_h, dims, 
                    BoxSize, h, redshift, NHI_values, NHI_bins)

            else:
                pos_h[start_h:start_h+offset+npart-begin]   = pos[begin-offset:]
                radii_h[start_h:start_h+offset+npart-begin] = radii[begin-offset:]
                MHI_h[start_h:start_h+offset+npart-begin]   = MHI[begin-offset:]
                start_h = start_h+offset+npart-begin
                begin = offset+npart

    return halo_mass, M_HI_tot, cross_section
"""

# This routine computes the DLAs cross-section of a single halo
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef DLAs_cross_section_1halo(np.float32_t[:,:] pos, np.float32_t[:] radii,
    np.float32_t[:] MHI, int dims, float BoxSize, float h, float redshift,
    np.float64_t[:] NHI_values, int NHI_bins):

    cdef int i, j, k, particles
    cdef float x_min, x_max, y_min, y_max, xW, yW, middle, BoxSize_region
    cdef double factor 
    cdef np.float64_t[:,:] NHI
    cdef np.float64_t[:] sigma, threshold

    particles = pos.shape[0]
    middle = BoxSize/2.0

    sigma     = np.zeros(NHI_bins, dtype=np.float64)
    threshold = np.zeros(NHI_bins, dtype=np.float64)

    # find the edges of the region
    x_min = np.min(pos[:,0]);  x_max = np.max(pos[:,0]);  xW = x_max-x_min
    y_min = np.min(pos[:,1]);  y_max = np.max(pos[:,1]);  yW = y_max-y_min

    if xW>BoxSize/2.0:
        for i in range(particles):
            if pos[i,0]<middle:  pos[i,0] += BoxSize
        x_min = np.min(pos[:,0]);  x_max = np.max(pos[:,0]);  xW = x_max-x_min
    
    if yW>BoxSize/2.0:
        for i in range(particles):
            if pos[i,1]<middle:  pos[i,1] += BoxSize
        y_min = np.min(pos[:,1]);  y_max = np.max(pos[:,1]);  yW = y_max-y_min

    BoxSize_region = max(yW, xW)

    # define array contining column densities
    NHI = np.zeros((dims,dims), dtype=np.float64)
    pos_arr = np.ascontiguousarray(pos[:,0:2])

    MASL.voronoi_RT_2D(np.asarray(NHI), pos_arr, 
        np.asarray(MHI), np.asarray(radii), x_min, y_min, 0, 1, BoxSize_region, 
        periodic=False, verbose=False)

    # convert (Msun/h)/(Mpc/h)^2 to cm^{-2}
    factor = h*(1.0+redshift)**2*\
        (UL.units().Msun_g)/(UL.units().mH_g)/(UL.units().Mpc_cm)**2
    for k in range(NHI_bins):
        threshold[k] = NHI_values[k]/factor

    for i in range(dims):
        for j in range(dims):
            for k in range(NHI_bins):
                if NHI[i,j]>threshold[k]:  sigma[k] += 1.0
                else:                      break
 
    for k in range(NHI_bins):
        sigma[k] = sigma[k]*BoxSize_region**2*1.0/(dims*dims)

    return np.asarray(sigma)



















import numpy as np 
import time,sys,os 
cimport numpy as np
cimport cython
#from cython.parallel import prange
from libc.math cimport sqrt,pow,sin,log10,abs
from libc.stdlib cimport malloc, free

############## AVAILABLE ROUTINES ##############
#pos_redshift_space
################################################



###############################################################################
#This routine receives the positions of the particles in configuration-space
#and return them in redshift-space
#pos --------------------> float array with the positions of the particles
#vel --------------------> float array with the velocities of the particles
#BoxSize ----------------> size of the simulation box
#Hubble -----------------> value of H(z) in (km/s)/(Mpc/h)
#redshift ---------------> redshift
#axis -------------------> axis along which perform the RSD
#The routines just uses: s = r + (1+z)/H(z)*v
#@cython.cdivision(True) never set this as % in python is different to c
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True) 
cpdef void pos_redshift_space(float[:,::1] pos, float[:,::1] vel,
                       float BoxSize, float Hubble, float redshift, int axis):

    cdef long particles,i
    cdef float factor

    particles = pos.shape[0]
    factor    = (1.0 + redshift)/Hubble

    for i in range(particles):
        pos[i,axis] = pos[i,axis] + vel[i,axis]*factor

        # neutrinos can cross the box multiple times
        while pos[i,axis]<0.0:
            pos[i,axis]+=BoxSize

        if pos[i,axis]>BoxSize:
            pos[i,axis] = (pos[i,axis])%BoxSize 
        
        #neutrinos can cross the box multiple times. Use % for the boundary
        #if pos[i,axis]>BoxSize or pos[i,axis]<0.0:
        #    pos[i,axis] = (pos[i,axis]+BoxSize)%BoxSize 
###############################################################################

################################ old routine ##################################
"""
def pos_redshift_space(pos,vel,BoxSize,Hubble,redshift,axis):
    #transform coordinates to redshift space
    delta_y=(vel[:,axis]/Hubble)*(1.0+redshift)  #displacement in Mpc/h
    pos[:,axis]+=delta_y #add distorsion to position of particle in real-space
    del delta_y

    #take care of the boundary conditions
    #beyond=np.where(pos[:,axis]>BoxSize)[0]; pos[beyond,axis]-=BoxSize
    #beyond=np.where(pos[:,axis]<0.0)[0];     pos[beyond,axis]+=BoxSize
    #del beyond

    # for neutrinos it could happen that in redshift-space their position
    # is further than 2 times size of the box, thus instead of doing
    # pos = pos+BoxSize or pos = pos - BoxSize we can just do pos = pos%BoxSize
    beyond = np.where((pos[:,axis]>BoxSize) | (pos[:,axis]<0.0))[0]
    pos[beyond,axis] = np.mod(pos[beyond,axis],BoxSize);  del beyond
"""
###############################################################################



from setuptools import setup
from setuptools.extension import Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize
import numpy


ext_modules = cythonize([

    Extension("integration_library", 
		["integration_library.pyx",'integration.c'],
		include_dirs=[numpy.get_include()],
		extra_compile_args=["-O3","-ffast-math","-march=native"]
		)])


setup(ext_modules = ext_modules)



# This simple script can be used as
# import numpy as np
# import cython_code as CC
# a = np.arange(123456789, dtype=np.int32)
# CC.pyprint()
# CC.pysum(a)
import numpy as np 
cimport numpy as np
import sys,os,time
cimport cython

cdef extern void cprint()
cdef extern double cc_sum(double *a, long elements)

# python function that calls a C function that just prints 'Hello'
def pyprint():
    cprint()

# python function that takes an array of doubles and computes the sum
# of their cubes by calling a C function
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def csum(double[::1] array):

    cdef double suma

    start = time.time()
    suma = cc_sum(&array[0], array.shape[0])
    print('\nTime taken C func = %.3f ms'%(1e3*(time.time()-start)))
    print('Sum = %.8e'%suma)


# python function that takes an array of doubles and computes the sum
# of their cubes by calling using cython
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def cysum(double[::1] array):

    cdef double suma=0.0
    cdef long i, elements=array.shape[0]

    start = time.time()
    for i in range(elements):
        suma += array[i]*array[i]*array[i]
    print('\nTime taken cython = %.3f ms'%(1e3*(time.time()-start)))
    print('Sum = %.8e'%suma)


# python function that takes an array of doubles and computes the sum
# of their cubes by calling using cython
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def pysum(array):

    start = time.time()
    suma = np.sum(array**3, dtype=np.float64)
    print('\nTime taken numpy = %.3f ms'%(1e3*(time.time()-start)))
    print('Sum = %.8e'%suma)



# This routine is designed to read in parallel Gadget snapshots.
# so far it can read POS, VEL and ID blocks. It automatically takes into 
# account whether IDs are uint32 or uint64

from mpi4py import MPI
import numpy as np
import readsnap
import sys,os
import time as Time

###### MPI DEFINITIONS ######                                    
comm=MPI.COMM_WORLD
nprocs=comm.Get_size()
myrank=comm.Get_rank()



def read_block(filename, block, parttype, physical_velocities=True,
               verbose=False):

    if parttype not in [0,1,2,3,4,5]:
        print('Routine can only read parttypes in [0,1,2,3,4,5,6]')
        sys.exit()

    # find the format,swap, total number of particles and subfiles
    head = readsnap.snapshot_header(filename)
    files, format, swap = head.filenum, head.format, head.swap
    nall, time, redshift = head.nall, head.time, head.redshift;  del head
    if myrank==0 and verbose:
        print('Reading snapshot with %d cores'%nprocs)
        print('Number of subfiles = %d'%files)

    # find the files each cpu reads
    subfiles = np.array_split(np.arange(files),nprocs)[myrank]

    # find the total number of particles to read
    Nall_local = np.int64(0)
    for i in subfiles:
        head = readsnap.snapshot_header(filename+'.%d'%i)
        Nall_local += head.npart[parttype]
    del head

    if verbose:
        print('core %03d reading %03d files: [%03d-%03d] %9d particles'\
            %(myrank,len(subfiles),subfiles[0],subfiles[-1],Nall_local))

    # check that all particles are read
    Nall = comm.reduce(Nall_local, op=MPI.SUM, root=0)
    if myrank==0 and Nall!=nall[parttype]:
        print('Read %d particles while expected %d'%(Nall,nall[parttype]))
        sys.exit()

    # find the data type
    if block=="POS ":  
        dt = np.dtype((np.float32,3))
        block_num = 2
    elif block=="VEL ":  
        dt = np.dtype((np.float32,3))
        block_num = 3
    elif block=="ID  ":  
        dt = np.uint32
        block_num = 4
    else:
        print('Block not found!');  sys.exit()

    # define the data array
    if myrank==0:  data = np.empty(Nall,       dtype=dt)
    else:          data = np.empty(Nall_local, dtype=dt)

    # do a loop over all subfiles
    offset_array = 0;  start = Time.time()
    for i in subfiles:

        # find subfile name and number of particles in it
        curfilename = filename+'.%d'%i
        head        = readsnap.snapshot_header(curfilename)
        npart       = head.npart 
        particles   = npart[parttype]

        offset_species = np.zeros(6,np.int64)
        allpartnum = np.int64(0)
        for j in range(6):
            offset_species[j] = allpartnum
            allpartnum += npart[j]

        # find the offset and the size of the block (for all particle types)
        offset_block,blocksize = readsnap.find_block(curfilename,format,swap,
                                                     block,block_num)

        # if long IDs change dt to np.uint64
        if i==subfiles[0] and block=="ID  ":
            if blocksize==np.dtype(dt).itemsize*allpartnum*2:
                dt = np.uint64

        # read file
        f = open(curfilename, 'rb')
        f.seek(offset_block + offset_species[parttype]*np.dtype(dt).itemsize, 
               os.SEEK_CUR)
        curdat = np.fromfile(f, dtype=dt, count=particles)
        f.close()

        if swap:  curdat.byteswap(True)

        data[offset_array:offset_array+particles] = curdat
        offset_array += particles
    if verbose:
        print('%d: Time to read files = %.2f'%(myrank,Time.time()-start))


    # slaves send master the particles read
    if myrank>0:
        comm.send(Nall_local, dest=0, tag=1) #number of particles read
        comm.Send(data,       dest=0, tag=2) #property read (pos,vel,ID..)
        return 0 #put 0 to avoid problems when reading pos and /1e3 #Mpc/h
    
    # master collect all information from slaves and return the array
    else:
        offset = Nall_local
        # do a loop over all slaves
        start = Time.time()
        for i in range(1,nprocs):
            npart = comm.recv(source=i, tag=1)
            comm.Recv(data[offset:offset+npart], source=i, tag=2)
            if verbose:
                print('Time to transfer files = %.2f'%(Time.time()-start))
            offset += npart

        if physical_velocities and block=="VEL " and redshift!=0:
            data *= np.sqrt(time)

        return data

